@@ -1,11 +1,7 @@
 import os
 import json
 from git import refresh
-from yaml import load, dump
-try:
-    from yaml import CLoader as Loader, CDumper as Dumper
-except ImportError:
-    from yaml import Loader, Dumper
+from ruamel.yaml import YAML
 import re
 from tangata import tangata_catalog_compile
 from functools import reduce
@@ -14,18 +10,19 @@ from whoosh.fields import *
 from whoosh.qparser import QueryParser, MultifieldParser
 from whoosh.filedb.filestore import RamStorage
 
-class CustomDumper(Dumper):
-    #Super neat hack to preserve the mapping key order. See https://stackoverflow.com/a/52621703/1497385
-    def represent_dict_preserve_order(self, data):
-        return self.represent_dict(data.items())
-    # def increase_indent(self, flow=False, indentless=False):
-    #     return super(MyDumper, self).increase_indent(flow, False)    
+# class CustomDumper(Dumper):
+#     #Super neat hack to preserve the mapping key order. See https://stackoverflow.com/a/52621703/1497385
+#     def represent_dict_preserve_order(self, data):
+#         return self.represent_dict(data.items())
+#     # def increase_indent(self, flow=False, indentless=False):
+#     #     return super(MyDumper, self).increase_indent(flow, False)    
 
-CustomDumper.add_representer(dict, CustomDumper.represent_dict_preserve_order)
+# CustomDumper.add_representer(dict, CustomDumper.represent_dict_preserve_order)
 
 skipDBTCompile = False
 disableRecompile = False
 lastGitIndex = {}
+tangataConfig = {}
 
 def setSkipDBTCompile(newSkipDBTCompile):
     global skipDBTCompile
@@ -35,12 +32,27 @@ def setDisableRecompile(newDisableRecompile):
     global disableRecompile
     disableRecompile = newDisableRecompile
 
+def setTangataConfig(newTangataConfig):
+    global tangataConfig
+    tangataConfig = newTangataConfig
+    tangata_catalog_compile.setTangataConfig(newTangataConfig)
+
 catalogPath = "./tangata_catalog.json"
 catalogIndexPath = "./tangata_catalog_index.json"
 catalog = {}
 catalogIndex = []
 catalogWhooshIndex = {}
 
+def loadSave():
+    global catalog
+    global catalogIndex
+    if os.path.exists("target/tangata_catalog.json") and os.path.exists("target/tangata_catalog_index.json"):
+        with open("target/tangata_catalog.json", "r") as cat:
+            catalog = json.load(cat)
+        with open("target/tangata_catalog_index.json", "r") as catIndex:
+            catalogIndex = json.load(catIndex)
+
+
 def refreshMetadata(sendToast):
     global catalog
     global catalogIndex
@@ -63,6 +75,10 @@ def refreshMetadata(sendToast):
     catalog = assemblingFullCatalog
     catalogIndex = assemblingCatalogIndex
     sendToast("Metadata has been refreshed successfully.", "success")
+    with open("target/tangata_catalog.json", "w") as cat:
+        json.dump(catalog, cat)
+    with open("target/tangata_catalog_index.json", "w") as catIndex:
+        json.dump(catalogIndex, catIndex)
 
 
 def searchModels(searchString):
@@ -107,7 +123,8 @@ def searchModels2(searchString):
                     "nodeID": thisMatch['nodeID'],
                     "modelName": catalog[thisMatch['nodeID']]['name'],
                     "modelDescription": catalog[thisMatch['nodeID']]['description'],
-                    "modelTags": catalog[thisMatch['nodeID']]['tags']
+                    "modelTags": catalog[thisMatch['nodeID']]['tags'],
+                    "promoteStatus": catalog[thisMatch['nodeID']]['promote_status']
                 })
         results = json.dumps(foundModels)
         return '{"results": ' + results + ',"searchString":"' + searchString + '"}'
@@ -115,21 +132,36 @@ def searchModels2(searchString):
 def get_model_tree():
     def filter_model_name(indexRecord):
         return indexRecord['type'] == "model_name"
+    def filter_sources(indexRecord):
+        return indexRecord['nodeID'].startswith("source.")
+    def filter_models(indexRecord):
+        return indexRecord['nodeID'].startswith("model.")
     def split_models(res, cur):
-        splitVal = reduce(lambda res, cur: {cur: res}, reversed(cur["nodeID"].split(".")), {})
+        splitVal = reduce(lambda res, cur: {cur: res}, reversed(catalog[cur["nodeID"]]['model_path'].split(".")[0].split("/")), {"nodeID": cur["nodeID"], "promote_status": catalog[cur["nodeID"]]['promote_status']})
+        res.append(splitVal)
+        return res
+    def split_sources(res, cur):
+        splitList = cur["nodeID"].replace("source.","sources.").split(".")
+        del splitList[1]
+        splitVal = reduce(lambda res, cur: {cur: res}, reversed(splitList), {"nodeID": cur["nodeID"], "promote_status": catalog[cur["nodeID"]]['promote_status']})
         res.append(splitVal)
         return res
     def merge_models(res, cur):
         return merge(res, cur)
 
-    all_models = list(filter(filter_model_name, catalogIndex))
+    all_models = list(filter(filter_models, filter(filter_model_name, catalogIndex)))
+    all_sources = list(filter(filter_sources, filter(filter_model_name, catalogIndex)))
+
     split_models = reduce(split_models, all_models, [])
-    resultObject = reduce(merge_models, split_models, {})
+    split_sources = reduce(split_sources, all_sources, [])
+    mergedModels = reduce(merge_models, split_models, {})
+    mergedSources = reduce(merge_models, split_sources, {})
+    resultObject = merge(mergedModels, mergedSources)
     return resultObject
     
 def get_db_tree():
     def get_db_keys(item):
-        db_keys = ["database", "schema", "name", "nodeID"]
+        db_keys = ["database", "schema", "name", "nodeID", "promote_status"]
         return {key: catalog[item][key] for key in db_keys}
 
     all_models = list(map(get_db_keys, catalog))
@@ -150,27 +182,50 @@ def findOrCreateMetadataYML(yaml_path, model_path, model_name, source_schema, mo
             else:
                 # createNewYML - source
                 newYAML = {"version": 2,"sources": [{"name": source_schema,"tables": [{"name": modelName}]}]}
-            yamlToWrite = dump(newYAML, Dumper=CustomDumper)
-            print(yamlToWrite)
             newYamlWrite = open(schemaPath, "w")
-            newYamlWrite.write(yamlToWrite)
+            yaml = YAML()
+            yaml.indent(mapping=2, sequence=4, offset=2)
+            yaml.dump(newYAML, newYamlWrite)
             return schemaPath
-        path = '' + model_path.replace('\\','/')
-        print(path)
-        print(path.rindex('/'))
-        print(path[0:path.rindex('/')])
+        path = '' + model_path
         path = path[0:path.rindex('/')]
         directory = os.path.dirname(path)
         if not os.path.exists(directory):
             # useSchemaYML - directory doesn't exist
             os.makedirs(directory)
-        schemaPath = path+'/schema.yml'
+        print(path)
+        print(os.listdir(path))
+        ymlsInFolder = len([file for file in os.listdir(path) if (file.endswith(".yml") or file.endswith(".yaml")) and os.path.isfile(path+"/"+file)])
+        
+        def singleYMLName():
+            for file in os.listdir(path):
+                    if (file.endswith(".yml") or file.endswith(".yaml")):
+                        return path+"/"+file
+            return None
+
+        def filePerFolder(defaultName):
+            if ymlsInFolder == 1:
+                existingYMLFile = singleYMLName()
+                if existingYMLFile:
+                    return existingYMLFile
+            return path+'/'+defaultName+'.yml'
+        if tangataConfig["schema_file_settings"] == "file_per_folder__folder_name":
+            schemaPath = filePerFolder(path[path.rindex('/')+1:])
+        elif tangataConfig["schema_file_settings"] == "file_per_folder__schema_yml":
+            schemaPath = filePerFolder("schema")
+        elif tangataConfig["schema_file_settings"] == "file_per_model__model_name":
+            schemaPath = path+'/'+model_name+'.yml'
+        else:
+            print("Issue with tangata config. Please report bug with the below:")
+            print(tangataConfig)
+
         try:
             if os.path.isfile(schemaPath):
                 # useSchemaYML - schemaPath exists
                 schemaPathRead = open(schemaPath, "r")
-                currentSchemaYML = load(schemaPathRead, Loader=Loader)
-                schemaPathRead.close()
+                yaml = YAML()
+                yaml.indent(mapping=2, sequence=4, offset=2)
+                currentSchemaYML = yaml.load(schemaPathRead)
                 if model_or_source == 'model':
                     # useSchemaYML - is model
                     if len(list(filter(lambda d: d['name'] == model_name, currentSchemaYML['models']))) > 0:
@@ -180,7 +235,7 @@ def findOrCreateMetadataYML(yaml_path, model_path, model_name, source_schema, mo
                         print('useSchemaYML - pushing model')
                         currentSchemaYML['models'].append({"name": model_name})
                         schemaPathWrite = open(schemaPath, "w")
-                        schemaPathWrite.write(dump(currentSchemaYML, Dumper=CustomDumper))
+                        yaml.dump(currentSchemaYML, schemaPathWrite)
                         schemaPathWrite.close()
                         print('useSchemaYML - pushed model')
                 else:
@@ -197,7 +252,7 @@ def findOrCreateMetadataYML(yaml_path, model_path, model_name, source_schema, mo
                             # pushing just table
                             list(filter(lambda d: d['name'] == source_schema, currentSchemaYML['sources']))['tables'].append({"name": model_name})
                         schemaPathWrite = open(schemaPath, "w")
-                        schemaPathWrite.write(dump(currentSchemaYML, Dumper=CustomDumper))
+                        yaml.dump(currentSchemaYML, schemaPathWrite)
                         schemaPathWrite.close()
                 return schemaPath
             else:
@@ -208,13 +263,15 @@ def findOrCreateMetadataYML(yaml_path, model_path, model_name, source_schema, mo
     print(model_name)
     if model_or_source == 'source':
         # is source
-        path = '' + model_path.replace('\\','/')
+        path = '' + model_path
         print(path)
         try:
             if os.path.isfile(path):
                 # first try path is file
                 pathRead = open(path, "r")
-                currentSchemaYML = load(pathRead, Loader=Loader)
+                yaml = YAML()
+                yaml.indent(mapping=2, sequence=4, offset=2)
+                currentSchemaYML = yaml.load(pathRead)
                 pathRead.close()
                 # opened yaml
                 if len(list(filter(lambda d: d['name'] == source_schema, currentSchemaYML['sources']))) > 0 and len(list(filter(lambda d: d['name'] == model_name, list(filter(lambda d: d['name'] == source_schema, currentSchemaYML['sources']))[0]['tables']))) > 0:
@@ -229,7 +286,7 @@ def findOrCreateMetadataYML(yaml_path, model_path, model_name, source_schema, mo
                         # pushing just table
                         list(filter(lambda d: d['name'] == source_schema, currentSchemaYML['sources']))['tables'].append({"name": model_name})
                     pathWrite = open(path, "w")
-                    pathWrite.write(dump(currentSchemaYML, Dumper=CustomDumper))
+                    yaml.dump(currentSchemaYML, pathWrite)
                     pathWrite.close()
                 return path
             else:
@@ -237,11 +294,13 @@ def findOrCreateMetadataYML(yaml_path, model_path, model_name, source_schema, mo
         except:
             return useSchemaYML()
     elif yaml_path is not None and len(yaml_path) > 0:
-        path = '' + yaml_path.replace('\\','/')
+        path = '' + yaml_path
         try:
             if os.path.isfile(path):
                 pathRead = open(path, "r")
-                currentSchemaYML = load(pathRead, Loader=Loader)
+                yaml = YAML()
+                yaml.indent(mapping=2, sequence=4, offset=2)
+                currentSchemaYML = yaml.load(pathRead)
                 pathRead.close()
                 if len(list(filter(lambda d: d['name'] == model_name, currentSchemaYML['models']))) > 0:
                     # found model in file
@@ -253,7 +312,7 @@ def findOrCreateMetadataYML(yaml_path, model_path, model_name, source_schema, mo
                     print('now pushed, list is now:')
                     print(currentSchemaYML)
                     pathWrite = open(path, "w")
-                    pathWrite.write(dump(currentSchemaYML, Dumper=CustomDumper))
+                    yaml.dump(currentSchemaYML, pathWrite)
                     pathWrite.close()
                 return path
             else:
@@ -282,7 +341,9 @@ def update_metadata(jsonBody, sendToast):
     if jsonBody['updateMethod'] == 'yamlModelProperty':
         schemaYMLPath = findOrCreateMetadataYML(jsonBody['yaml_path'], jsonBody['model_path'], jsonBody['model'], jsonBody['node_id'].split(".")[2], jsonBody['node_id'].split(".")[0])
         schemaPathRead = open(schemaYMLPath, "r")
-        currentSchemaYML = load(schemaPathRead, Loader=Loader)
+        yaml = YAML()
+        yaml.indent(mapping=2, sequence=4, offset=2)
+        currentSchemaYML = yaml.load(schemaPathRead)
         schemaPathRead.close()
         if jsonBody['node_id'].split(".")[0] == 'model':
             currentSchemaYMLModel = list(filter(lambda d: d['name'] == jsonBody['model'], currentSchemaYML['models']))[0]
@@ -290,43 +351,49 @@ def update_metadata(jsonBody, sendToast):
             currentSchemaYMLModel = list(filter(lambda d: d['name'] == jsonBody['model'], list(filter(lambda d: d['name'] == jsonBody['node_id'].split(".")[2], currentSchemaYML['sources']))[0]['tables']))[0]
         currentSchemaYMLModel[jsonBody['property_name']] = jsonBody['new_value']
         pathWrite = open(schemaYMLPath, "w")
-        pathWrite.write(dump(currentSchemaYML, Dumper=CustomDumper))
+        yaml.dump(currentSchemaYML, pathWrite)
         pathWrite.close()
     elif jsonBody['updateMethod'] == 'yamlModelTags':
         if jsonBody['node_id'].split(".")[0] == 'model':
             dbtProjectYMLModelPath = ['models', jsonBody['node_id'].split(".")[1]]
-            splitModelPath = jsonBody['model_path'].split(".")[0].split("\\")
+            splitModelPath = jsonBody['model_path'].split(".")[0].split("/")
             splitModelPath.pop(0)
             dbtProjectYMLModelPath = dbtProjectYMLModelPath + splitModelPath
             readDbtProjectYml = open(''+"dbt_project.yml", "r")
-            dbtProjectYML = load(readDbtProjectYml, Loader=Loader)
+            yaml = YAML()
+            yaml.indent(mapping=2, sequence=4, offset=2)
+            dbtProjectYML = yaml.load(readDbtProjectYml)
             readDbtProjectYml.close()
-            jsonToInsert = ""
+            jsonTags = json.loads("[\""+"\",\"".join(jsonBody['new_value'])+"\"]")
+            yamlModel = dbtProjectYML
             for pathStep in dbtProjectYMLModelPath:
-                jsonToInsert += "{\"" + pathStep + "\": "
-            jsonToInsert += "{\"tags\": [\""+"\",\"".join(jsonBody['new_value'])+"\"]}"
-            for pathStep in dbtProjectYMLModelPath:
-                jsonToInsert += "}"
-            jsonToInsert = json.loads(jsonToInsert)
-            newDBTProjectYML = merge(dbtProjectYML, jsonToInsert)
-            writeDbtProjectYml = open(''+"dbt_project.yml", "w")
-            writeDbtProjectYml.write(dump(newDBTProjectYML, Dumper=CustomDumper))
+                if not pathStep in yamlModel:
+                    yamlModel[pathStep] = {'tags': []}
+                yamlModel = yamlModel[pathStep]
+
+            yamlModel['tags'] = jsonTags
+            writeDbtProjectYml = open("dbt_project.yml", "w")
+            yaml.dump(dbtProjectYML, writeDbtProjectYml)
             writeDbtProjectYml.close()
 
         else:
             schemaYMLPath = findOrCreateMetadataYML(jsonBody['yaml_path'], jsonBody['model_path'], jsonBody['model'], jsonBody['node_id'].split(".")[2], jsonBody['node_id'].split(".")[0])
             schemaPathRead = open(schemaYMLPath, "r")
-            currentSchemaYML = load(schemaPathRead, Loader=Loader)
+            yaml = YAML()
+            yaml.indent(mapping=2, sequence=4, offset=2)
+            currentSchemaYML = yaml.load(schemaPathRead)
             schemaPathRead.close()
             currentSchemaYMLModel = list(filter(lambda d: d['name'] == jsonBody['model'], list(filter(lambda d: d['name'] == jsonBody['node_id'].split(".")[2], currentSchemaYML['sources']))[0]['tables']))[0]
             currentSchemaYMLModel['tags'] = jsonBody['new_value']
             pathWrite = open(schemaYMLPath, "w")
-            pathWrite.write(dump(currentSchemaYML, Dumper=CustomDumper))
+            yaml.dump(currentSchemaYML, pathWrite)
             pathWrite.close()
     elif jsonBody['updateMethod'] == 'yamlModelColumnProperty':
         schemaYMLPath = findOrCreateMetadataYML(jsonBody['yaml_path'], jsonBody['model_path'], jsonBody['model'], jsonBody['node_id'].split(".")[2], jsonBody['node_id'].split(".")[0])
         schemaPathRead = open(schemaYMLPath, "r")
-        currentSchemaYML = load(schemaPathRead, Loader=Loader)
+        yaml = YAML()
+        yaml.indent(mapping=2, sequence=4, offset=2)
+        currentSchemaYML = yaml.load(schemaPathRead)
         schemaPathRead.close()
         if jsonBody['node_id'].split(".")[0] == 'model':
             currentSchemaYMLModel = list(filter(lambda d: d['name'] == jsonBody['model'], currentSchemaYML['models']))[0]
@@ -343,12 +410,14 @@ def update_metadata(jsonBody, sendToast):
             currentSchemaYMLModelColumn = list(filter(lambda d: d['name'] == jsonBody['column'], currentSchemaYMLModel['columns']))[0]
         currentSchemaYMLModelColumn[jsonBody['property_name']] = jsonBody['new_value']
         pathWrite = open(schemaYMLPath, "w")
-        pathWrite.write(dump(currentSchemaYML, Dumper=CustomDumper))
+        yaml.dump(currentSchemaYML, pathWrite)
         pathWrite.close()
     elif jsonBody['updateMethod'] == 'yamlModelColumnTest':
         schemaYMLPath = findOrCreateMetadataYML(jsonBody['yaml_path'], jsonBody['model_path'], jsonBody['model'], jsonBody['node_id'].split(".")[2], jsonBody['node_id'].split(".")[0])
         schemaPathRead = open(schemaYMLPath, "r")
-        currentSchemaYML = load(schemaPathRead, Loader=Loader)
+        yaml = YAML()
+        yaml.indent(mapping=2, sequence=4, offset=2)
+        currentSchemaYML = yaml.load(schemaPathRead)
         schemaPathRead.close()
         if jsonBody['node_id'].split(".")[0] == 'model':
             currentSchemaYMLModel = list(filter(lambda d: d['name'] == jsonBody['model'], currentSchemaYML['models']))[0]
@@ -370,7 +439,7 @@ def update_metadata(jsonBody, sendToast):
             del currentSchemaYMLModelColumn['tests']
         print(currentSchemaYMLModel)
         pathWrite = open(schemaYMLPath, "w")
-        pathWrite.write(dump(currentSchemaYML, Dumper=CustomDumper))
+        yaml.dump(currentSchemaYML, pathWrite)
         pathWrite.close()
     return "success"
 
